// nmea-mtk.cpp : Defines the entry point for the console application.
//

#undef _UNICODE
#undef UNICODE

#include "stdafx.h"

int Send_NMEA(HANDLE hCom, DCB dcb, char *cmd)
{
	char buffer[1024], *b = buffer;
	int len = strlen(cmd);
	if (len > 1000) len = 1000;
	if (cmd[0] != '$')
		*b++ = '$';
	strcpy_s(b, sizeof(buffer)-2, cmd);
	len = strlen(buffer);
	if (len < 4 || (buffer[len-2] != '*')) {
		if (buffer[len-1] != '*')
			buffer[len++] = '*';
		unsigned xor_sum = 0;
		for (int i = 1; i < len-1; i++)
			xor_sum ^= buffer[i];
		sprintf_s(buffer + len, sizeof(buffer)-len, "%02X\r\n", xor_sum);
		len += 4;
	}
	DWORD bytesWritten;
	WriteFile(hCom, buffer, len, &bytesWritten, NULL);
	printf("%s", buffer);
	return bytesWritten == len;
}

int Send_MTK(HANDLE hCom, DCB dcb, char *cmd)
{
   // Start monitoring loop, to wait for a lull:
   DWORD ft, nt, lt;

   ft = lt = nt = GetTickCount();
   char last = '\0';
   DWORD bytesRead, bytesWritten;
   printf("\nWaiting for idle time\n");
   do {
	   char buf[4096];
	   ReadFile(hCom, buf, sizeof(buf), &bytesRead, NULL);
	   if (bytesRead) {
		   for (DWORD b = 0; b < bytesRead; b++) {
			   if (buf[b] >= ' ')
				   printf("%c", buf[b]);
			   else {
				   if (buf[b] == '\r')
					   printf("<CR>");
				   else if (buf[b] == '\n')
					   printf("<LF>\n");
				   else
					   printf("<0x%02x>", buf[b]);
			   }
		   }
		   last = buf[bytesRead-1];
		   lt = GetTickCount();
	   }
	   else {
		   nt = GetTickCount();
		   if (last == '\n' && (nt-lt > 50))
			   break;
	   }
   } while (nt - ft < 1000); // Try for max 1 second

   if (nt-ft >= 1000)
	   printf("\nWaited 1 s, no 50 ms gap found\n");
   else
	   printf("\nFound a 50 ms opening\n");

   WriteFile(hCom, "0", 1, &bytesWritten, NULL);
   DWORD ms;
   ft = nt;
   // Wait up to 200 ms for CD to turn off:
   while ((GetCommModemStatus(hCom, &ms) && (ms & MS_RLSD_ON)) || 
			(GetTickCount() - ft < 200)) {
		printf("%s\r", ms & MS_RLSD_ON ? "CD_on":"CD_off");
   }
   // CD still on?
   if (!(ms & MS_RLSD_ON))
		printf("\nCD is now off, ready to send cmd\n");

   return Send_NMEA(hCom, dcb, cmd);
}

void PrintCommState(DCB dcb)
{
    //  Print some of the DCB structure values
    printf( "\nBaudRate = %d, ByteSize = %d, Parity = %d, StopBits = %d\n", 
              dcb.BaudRate, 
              dcb.ByteSize, 
              dcb.Parity,
              dcb.StopBits );
}

int Syntax(char *msg)
{
	if (msg)
		fprintf(stderr, "%s", msg);
	fprintf(stderr, 
		"nmea-mtk (c) 2011 Terje Mathisen\n"\
		"Syntax: nmea-mtk [options]\n"\
		"\n"\
		"Default with no options is to enable WAAS and GGA\n"\
		"\n"\
		"Options:\n"\
		" -b baud (default 9600)\n"\
		" -p port (default \\\\.\\COM1)\n"\
		" -c command (send $command*xxCRLF to gps, can be repeated)\n"\
		" -t seconds (echo all NMEA sentences for t seconds\n"\
		"\n"\
		"Documented commands and responses:\n"\
		"\n"\
		"$PMTK000*     : Test packet\n"\
		"$PMTK010,001* : Output system message\n"\
		"$PMTK101*     : Hot restart\n"\
		"$PMTK102*     : Warm restart\n"\
		"$PMTK103*     : Cold restart\n"\
		"$PMTK104*     : Factory reset\n"\
		"\n"\
		"$PMTK251,4800*: Set baud rate (0,4800,9600,19200,38400,57600,115200)\n"\
		"\n"\
		"$PMTK300,307* : Set position fix interval (307 ms is the shortest on my Sure board)\n"\
		"$PMTK400*     : Query Position fix interval\n"\
		"$PMTK500,n*   : Current Position fix interval (in ms)\n"\
		"\n"\
		"$PMTK301,2*   : Set DGPS mode (0:off, 1:RTCM, 2:WAAS)\n"\
		"$PMTK401*     : Query DGPS mode\n"\
		"$PMTK501,n*   : Current DGPS mode status\n"\
		"\n"\
		"$PMTK313,1*   : Enable SBAS (0:disable)\n"\
		"$PMTK413*     : Query SBAS\n"\
		"$PMTK513      : Current SBAS Status\n"\
		"\n"\
		"$PMTK314,-1*  : Turn on default NMEA sentences\n"\
		"$PMTK314,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0* : Turn on GGA only\n"\
		"    The sentence order is GLL, RMC, VTG, GGA, GSA, GSV, GRS, GST,\n"\
		"    (5 undocumented sentences), MALM, MEPH, MDGP, MDBG, ZDA, MCHN\n"\
		"$PMTK414*     : Query NMEA sentences\n"\
		"$PMTK514      : Current NMEA sentences\n"\
		"\n"\
		"$PMTK320,0*   : Turn off (0) or on (1) power saving mode (test only!)\n"\
		"$PMTK420*     : Query power saving mode\n"\
		"$PMTK520,n*   : Current Power saving mode\n"\
		"\n"\
		"$PMTK330,0*   : Set Datum (0 = WGS84)\n"\
		"$PMTK430*     : Query Datum\n"\
		"$PMTK530,n*   : Current Datum used\n"\
		"\n"\
		"$PMTK331,6377397.155,299.1528128,-148.0,507.0,685.0* : Set custom datum\n"\
		"    (semi-major axis,eccentr.,WGS84 dx offset, dy, dz)\n"\
		"$PMTK431*     : Query Custom Datum\n"\
		"$PMTK531,n*   : Current Custom Datum used\n"\
		"\n"\
		"$PMTK390,n*     : Set Flash settings: write_count,hz,baud,\n"\
		"    GLL,RMC,VTG,GSA,GSV,GGA,ZDA,MCHN (default NMEA sentences to output)\n"\
		"$PMTK490,n*     : Query Flash settings\n"\
		"$PMTK590,n*     : Current Flash settings\n"\
		"\n"\
		"$PMTK604*     : Query FW version\n"\
		"$PMTK704,v1,v2,v3*: FW version numbers\n"\
		"$PMTK605*     : Query FW release info\n"\
		"$PMTK705,releaseStr,ModelID*: FW release\n");
	return 1;
}

int nval(char *a)
{
	int v = 0;
	char c;
	while ((c = *a++) != '\0') {
		if (c >= '0' && c <= '9')
			v = v*10 + c-'0';
	}
	return v;
}

int main( int argc, char *argv[] )
{
   DCB dcb;
   HANDLE hCom;
   BOOL fSuccess;
   char *pcCommPort = "\\\\.\\COM1"; // Using this syntax allows COM1 to COM255
   DWORD baud = 9600;
   DWORD echotime = 4;   // Default to echo NMEA for 4 seconds
   char *command_to_send = NULL;

   // Check any command line options:
   for (int a = 1; a < argc; a++) {
	   char *arg = argv[a];
	   if (arg[0] == '-') {
		   a++;
		   if (arg[1] == 'b') {
			   baud = nval(argv[a]);
		   }
		   else if (arg[1] == 'p') {
			   pcCommPort = argv[a];
		   }
		   else if (arg[1] == 't') {
			   echotime = nval(argv[a]);
		   }
		   else if (arg[1] == 'c') {
			   command_to_send = argv[a];
		   }
		   else
			   return Syntax(NULL);
	   }
	   else
		   return Syntax(NULL);
   }

   //  Open a handle to the specified com port.
   hCom = CreateFile( pcCommPort,
                      GENERIC_READ | GENERIC_WRITE,
                      0,      //  must be opened with exclusive-access
                      NULL,   //  default security attributes
                      OPEN_EXISTING, //  must use OPEN_EXISTING
                      0,//  overlapped I/O
                      NULL ); //  hTemplate must be NULL for comm devices

   if (hCom == INVALID_HANDLE_VALUE) 
   {
       //  Handle the error.
       printf ("CreateFile failed with error %d.\n", GetLastError());
       return (1);
   }

   //  Initialize the DCB structure.
   SecureZeroMemory(&dcb, sizeof(DCB));
   dcb.DCBlength = sizeof(DCB);

   //  Build on the current configuration by first retrieving all current
   //  settings.
   fSuccess = GetCommState(hCom, &dcb);

   if (!fSuccess) 
   {
      //  Handle the error.
      printf ("GetCommState failed with error %d.\n", GetLastError());
      return (2);
   }

   //  Fill in some DCB values and set the com state: 
   //  baud bps, 8 data bits, no parity, and 1 stop bit.
   dcb.BaudRate = baud;     //  baud rate
   dcb.ByteSize = 8;             //  data size, xmit and rcv
   dcb.Parity   = NOPARITY;      //  parity bit
   dcb.StopBits = ONESTOPBIT;    //  stop bit

   fSuccess = SetCommState(hCom, &dcb);

   if (!fSuccess) 
   {
      //  Handle the error.
      printf ("SetCommState failed with error %d.\n", GetLastError());
      return (3);
   }

   PrintCommState(dcb);       //  Output to console

   COMMTIMEOUTS tout;
   memset(&tout, 0, sizeof(tout));
   tout.ReadIntervalTimeout = MAXDWORD;
   SetCommTimeouts(hCom, &tout);

   if (command_to_send) { // Send all found commands!
	   for (int a = 1; a < argc; a++) {
		   char *arg = argv[a];
		   if (arg[0] == '-') {
			   a++;
			   if (arg[1] == 'c') {
				   command_to_send = argv[a];
				   Send_MTK(hCom, dcb, command_to_send);
			   }
		   }
	   }
   }
   else {
	  // Enable SBAS diff
	  Send_MTK(hCom, dcb, "$PMTK301,2*");
	  // Turn on WAAS
	  Send_MTK(hCom, dcb, "$PMTK313,1*");

	  // Turn on GGA, turn all others off
	  Send_MTK(hCom, dcb, "$PMTK314,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0*");
	  //Send_MTK(hCom, dcb, "$PMTK314,-1*");
   }

   // Echo all data for the next 4 seconds:
   DWORD ft = GetTickCount();
   do {
	   DWORD bytesRead;
	   char buf[4096];
	   ReadFile(hCom, buf, sizeof(buf), &bytesRead, NULL);
	   if (bytesRead) {
		   for (DWORD b = 0; b < bytesRead; b++) {
			   if (buf[b] >= ' ')
				   printf("%c", buf[b]);
			   else {
				   if (buf[b] == '\r')
					   printf("<CR>");
				   else if (buf[b] == '\n')
					   printf("<LF>\n");
				   else
					   printf("<0x%02x>", buf[b]);
			   }
		   }
	   }
	   Sleep(1);
   } while (GetTickCount() - ft < (echotime*1000));
   CloseHandle(hCom);

   printf ("NMEA-MTK finished.\n", pcCommPort);
   return (0);

}